/*
** Module   : IndexSelectionWorker.java
** Abstract : provides access to schema index data and helpers to calc the 
**            actual implicit index in use
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20051013   @23061 Provides access to schema index data and
**                           helpers to calc the actual implicit index in
**                           use.
** 002 GES 20051020   @23078 Major rewrite to support more complex rules
**                           and more complex tie breaks.
** 003 GES 20051020   @23084 Added sort and index overlap calculation.
** 004 GES 20051020   @23086 Added simple order by generation.
** 005 GES 20051026   @23100 Added import for persist. Updated order-by
**                           generation to use fully qualified field
**                           names.
** 006 ECF 20060202   @24257 Added support for temp/work-table schema
**                           data lookups. Implemented visitAst method to
**                           load current, temporary schema data into
**                           P2OLookup's state each time a new source file
**                           is visited during conversion.
** 007 ECF 20060406   @25392 Fixed order by clauses for buffers. The
**                           backing table's name was always being used as
**                           the field qualifier, even in cases where an
**                           explicitly named buffer was in use.
** 008 ECF 20060509   @26041 Added findBeginsIndexMatch() method. Testing
**                           indicates that Progress actually gives begins
**                           matches precedence over range matches.
**                           Previously, these were given equal weight.
** 009 ECF 20060824   @28871 Added support for by clauses which use extent
**                           field elements. This changes the content of
**                           the resulting, HQL order by clauses.
** 010 ECF 20080721   @39140 Minor optimization. Use StringBuilder instead
**                           of StringBuffer.
** 011 ECF 20090306   @41433 Fixed getIndex(). Use case-insensitive test
**                           when finding a P2JIndex for a given name,
**                           since that name may have been provided by the
**                           original code via USE-INDEX, and its case may
**                           not match the name stored in a P2JIndex.
** 012 SVL 20090505   @42094 mergeOrderByCriteria() added. Sort components
**                           can be matched in the reverse way.
** 013 GES 20090518   @42380 Import change.
** 014 SVL 20091003   @44104 Added getIndexNameByPrefix() function.
** 015 SVL 20110813          Fixed leadingSortComponentsMatch: reverse match
**                           only of the first component should trigger an
**                           attempt to check the possibility of reverse
**                           match of the whole sort string.
** 016 VMN 20140105          Added support for custom denormalization of fields with extent.
*/

package com.goldencode.p2j.convert;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.pattern.*;


/**
 * A pattern worker whose purpose is to provide helpers for the calculation
 * of all implicit index selections (based on record phrase and sorting
 * criteria analysis).  This worker also provides rule-set access to the 
 * schema index data.
 * <p>
 * The rules by which this index selection is handled are very complex.
 * Please see the <a href="../schema/package.html">Schema Package Summary</a>
 * for complete details.
 */
public class IndexSelectionWorker
extends AbstractPatternWorker
implements ProgressParserTokenTypes
{
   /** Schema name for the record being processed. */
   private String schemaname = null;
   
   /**
    * Stores the list of the equality matches for the current record phrase.
    */ 
   private Set equalityMatches = new HashSet();
   
   /**
    * Stores the list of the begins matches for the current record phrase.
    */ 
   private Set beginsMatches = new HashSet();
   
   /**
    * Stores the list of the range matches for the current record phrase.
    */ 
   private Set rangeMatches = new HashSet();
   
   /**
    * Stores the list of the sort matches for the current record phrase in
    * the order of increasing specificity.
    */ 
   private Map sortMatches = new LinkedHashMap();
   
   /**
    * Stores the list of the word matches for the current record phrase.
    */ 
   private Set wordMatches = new HashSet();
   
   /** Flag to control whether range processing is necessary. */
   private boolean rangeBypass = false;
   
   /**
    * Default constructor which initializes libraries.
    */
   public IndexSelectionWorker()
   {
      super();
      setLibrary(new IndexHelper());
   }
   
   /**
    * This method is called each time a new AST is visited, just before the
    * pattern engine processes that AST.  We exploit this hook by loading the
    * current temp/work-table schema (P2O) AST into the <code>P2OLookup</code>
    * class' state.  This enables us to resolve references to temp and work
    * tables, which are local to the currently loaded source file AST.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
      P2OLookup.loadTemporarySchema(ast.getFilename());
   }
   
   /**
    * Obtains an iterator to the current record's list of indexes.
    *
    * @return   The list of indexes.
    */
   private Iterator indexes()
   {
      Iterator iter = null;
      
      try
      {
         iter = P2OLookup.indexes(schemaname);   
      }
      
      catch (SchemaException se)
      {
         se.printStackTrace();
      }
      
      return iter;
   }
   
   /**
    * Provides the core user-functions exported to pattern engine rules for
    * Progress index selection calculation and schema index access.
    */
   public class IndexHelper
   {      
      /**
       * Initialize data structures and load schema data related to
       * processing the specified record name.  This should be called
       * upon encountering (walking) a new record phrase.
       *
       * @param    schemaname
       *           The fully qualified schema name of the record being
       *           processed.
       */
      public void initRecordPhrase(String schemaname)
      {
         // save schema name
         IndexSelectionWorker.this.schemaname = schemaname;
         
         // initialize match maps
         equalityMatches.clear();
         beginsMatches.clear();
         rangeMatches.clear();
         sortMatches.clear();
         wordMatches.clear();
      }
      
      /**
       * Use the recorded index component matches to determine the implicitly
       * selected index for this record phrase.  This should be called
       * upon exiting (ascending back to) a record phrase node.
       * <p>
       * The following is the precedence order:
       * <ol>
       *    <li> A unique index with all components in equality matches.  
       *    <li> A word index matched via contains.
       *    <li> The index with the most sequential leading equality matches.
       *    <li> The index with the most sequential leading range matches.
       *    <li> The index with the most sequential leading sort matches.
       *    <li> The primary index.
       * </ol>
       * <p>
       * See <a href="../schema/package.html">Schema Package Summary</a>
       * for complete details.
       *
       * @return   The name of the index selected or <code>""</code> (the
       *           empty string) in the case where there are no indexes
       *           defined for this table and thus there is no primary
       *           index to use as a default.
       */
      public String calcIndexSelection()
      {
         List   selectable = null;
         String result     = null;
         
         rangeBypass = false;
         
         do
         {
            // unique index match 
            result = findUniqueIndexMatch();
            if (result != null) break;
            
            // word index match
            result = findWordIndexMatch();
            if (result != null) break;
            
            // restart with all possible indexes
            selectable = getFullIndexList();
            
            // filter down by most sequential leading equality matches
            selectable = findLeadingIndexMatch(selectable);
                        
            if (selectable.size() > 0)
            {            
               // exit if we found a single match
               result = readSingleIndex(selectable);
               if (result != null) break;               
               
               // have we already processed ranges sufficiently?
               if (!rangeBypass)
               {
                  // filter down by first index component "begins" matches
                  selectable = findBeginsIndexMatch(selectable);
                  
                  // exit if we found a single match
                  result = readSingleIndex(selectable);
                  if (result != null) break;
                  
                  // filter down by first index component range matches
                  selectable = findRangeIndexMatch(selectable);
                  
                  // exit if we found a single match
                  result = readSingleIndex(selectable);
                  if (result != null) break;               
               }
               
               // most sequential leading sort matches
               result = findSortIndexMatch(selectable);
               if (result != null) break;               
            }
            
            // primary index
            P2JIndex primary = getPrimaryIndex();
            if (primary != null)
            {
               // default choice is the explicitly defined primary index
               result = primary.getName();
            }
            else
            {
               // the implicitly defined primary index (just recid)
               result = "";
            }
         }
         while (false);
         
         return result;
      }
      
      /**
       * Compares the given selected index with the stored sort criteria to 
       * determine if the the complete sort criteria is an exact match or a
       * compatible subset match to the selected index.  In this case, this
       * method will return <code>true</code>.  Any incompatibility between
       * the selected index and the sort criteria will generate a return of 
       * <code>false</code>.
       * <p>
       * All sort criteria (the list of index components that are sort
       * matches) must match in name, specificity (order of occurrance) and
       * sort direction.  If any of the sort matches does not exist in the  
       * selected index, exists in a different sequential position or differs
       * in direction (ascending/descending), then this is not compatible.
       * <p>
       * If the selected index has more index components than the sort
       * criteria AND all the sort criteria are a leading sequential subset,
       * this is still an overlap.  At this time, this routine will consider
       * any gaps as incompatibilities (it is unknown if this is different
       * from the Progress compiler's approach).
       *
       * @param    selected
       *           The name of the index to compare against. 
       *
       * @return   A value greater than <code>0</code> if the previously
       *           recorded sort criteria are compatible with all or a leading
       *           subset of the selected index in the direct order. A value
       *           less than <code>0</code> if the previously recorded sort
       *           criteria are compatible with all or a leading subset of the
       *           selected index, but in the reverse order. <code>0</code> in
       *           any other condition, including the case where there is a
       *           <code>null</code>, empty string or invalid index name 
       *           provided as the selected index.
       */
      public int calcSortOverlap(String selected)
      {
         P2JIndex idx = getIndex(selected);
         
         if (idx == null)
         {
            return 0;
         }
         
         int num = leadingSortComponentsMatch(idx);

         if (num != 0 && num == sortMatches.size())
         {
            return 1;
         }
         else if (num != 0 && -num == sortMatches.size())
         {
            return -1;
         }
         else
         {
            return 0;
         }
      }
      
      /**
       * Converts the given index into an equivalent order by statement. If 
       * no index is given, the empty string will be returned.
       *
       * @param    selected
       *           The name of the index from which to generate the order by
       *           clause.
       * @param    bufName
       *           Name of the Java variable which represents the buffer.
       * @param    reverse
       *           If <code>true</code> then the index should be taken in the
       *           reverse order. 
       *
       * @return   The generated order by string or the empty string if
       *           the empty string or an invalid index name is provided as
       *           the selected index.
       */
      public String calcOrderByIndex(String selected,
                                     String bufName,
                                     boolean reverse)
      throws SchemaException
      {
         StringBuilder result = new StringBuilder("");
         String        field  = null;
         
         P2JIndex idx = getIndex(selected);
         
         if (idx != null)
         {
            Iterator comps = idx.components();
            
            P2JIndexComponent nextComp = null;
            
            while (comps.hasNext())
            {
               if (result.length() > 0)
               {
                  result.append(", ");
               }
               
               nextComp = (P2JIndexComponent) comps.next();
               String sname = nextComp.getName();
               field = P2OLookup.javaPropertyName(sname, false, null);
               
               result.append(bufName);
               result.append(".");
               result.append(field);                                                 
               result.append(" ");
               result.append((reverse ? !nextComp.isDescending() :
                                        nextComp.isDescending()) ?
                                           "desc" : "asc");
            }
         }
            
         return result.toString();
      }
      
      /**
       * Converts the sort criteria into an equivalent order by statement. If 
       * no sort matches exist, the empty string will be returned.
       * <p>
       * The string keys of both the <code>criteria</code> map (or the
       * internal map of sort matches, if <code>criteria</code> is
       * <code>null</code>) and the <code>bufNames</code> map will end with
       * an array dereferencing trailer (e.g., "[0]"), IF the associated sort
       * component refers to a specific element within an extent field.  This
       * trailer already will be adjusted to assume a 0-based array.
       *
       * @param    criteria
       *           An ordered list of the sort criteria with the key as the
       *           fully qualified schema name of the field and the value
       *           as a <code>Boolean</code> that specifies that the sort
       *           direction is descending if <code>true</code>, otherwise
       *           that field will have an ascending sort direction.  The
       *           list MUST be ordered in the same order (from least
       *           granular to most granular) of the original BY clauses.
       *           If <code>null</code>, the previously recorded sort
       *           match criteria will be used.
       * @param    bufNames
       *           Map of fully qualified schema names of fields to the Java
       *           names of the buffers which they reference.
       *
       * @return   The generated order by string or the empty string if
       *           no sort criteria exist.
       */
      public String calcOrderByCriteria(LinkedHashMap criteria, Map bufNames)
      throws SchemaException
      {
         StringBuilder result = new StringBuilder("");
         String        field  = null;
         
         Map      matches = (criteria == null ? sortMatches : criteria);
         Iterator sort    = matches.keySet().iterator();
         
         while (sort.hasNext())
         {
            if (result.length() > 0)
            {
               result.append(", ");
            }
            
            String  critName = (String) sort.next();
            int pos = critName.indexOf("[");
            String sname = (pos < 0 ? critName : critName.substring(0, pos));
            Boolean descend  = (Boolean) matches.get(critName);
            
            result.append(bufNames.get(critName));
            result.append(".");
            field = P2OLookup.javaPropertyName(sname, false, null);
            result.append(field);
            if (pos > 0)
            {
               result.append(critName.substring(pos));
            }
            result.append(" ");
            result.append(descend.booleanValue() ? "desc" : "asc");
         }
            
         return result.toString();
      }

      /**
       * Merges the sort criteria and the selected index into an equivalent
       * order by statement. If no sort matches exist and the index is
       * <code>null</code>, the empty string will be returned.
       * <p>
       * Merging is performed in the following way: first, sort criteria are
       * emitted; then the sort criteria specified by index which are not
       * already present in the emitted criteria (into direct or reverse way)
       * are appended to the output. 
       * <p>
       * The string keys of both the <code>criteria</code> map (or the
       * internal map of sort matches, if <code>criteria</code> is
       * <code>null</code>) and the <code>bufNames</code> map will end with
       * an array dereferencing trailer (e.g., "[0]"), IF the associated sort
       * component refers to a specific element within an extent field.  This
       * trailer already will be adjusted to assume a 0-based array.
       *
       * @param    criteria
       *           An ordered list of the sort criteria with the key as the
       *           fully qualified schema name of the field and the value
       *           as a <code>Boolean</code> that specifies that the sort
       *           direction is descending if <code>true</code>, otherwise
       *           that field will have an ascending sort direction.  The
       *           list MUST be ordered in the same order (from least
       *           granular to most granular) of the original BY clauses.
       *           If <code>null</code>, the previously recorded sort
       *           match criteria will be used.
       * @param    bufNames
       *           Map of fully qualified schema names of fields to the Java
       *           names of the buffers which they reference.
       * @param    selected
       *           The name of the index which is merged with the order by
       *           clause. Can be <code>null</code>.
       * @param    indexBufName
       *           Name of the Java variable which represents the buffer for
       *           the specified index.
       *
       * @return   The generated order by string or the empty string if
       *           no sort criteria exist and the index is <code>null</code>.
       */
      public String mergeOrderByCriteria(LinkedHashMap criteria,
                                         Map bufNames,
                                         String selected,
                                         String indexBufName)
      throws SchemaException
      {
         Map    byElements = (criteria == null ? sortMatches : criteria);

         String byPart = calcOrderByCriteria((LinkedHashMap) byElements,
                                                             bufNames);
         StringBuilder result = new StringBuilder(byPart);

         P2JIndex idx = getIndex(selected);

         if (idx != null)
         {
            Iterator comps = idx.components();

            P2JIndexComponent nextComp = null;

            while (comps.hasNext())
            {
               nextComp = (P2JIndexComponent) comps.next();
               String sname = nextComp.getName();

               if (byElements.get(sname) != null)
                  continue;

               String field = P2OLookup.javaPropertyName(sname, false, null);
                                                  
               if (result.length() > 0)
               {
                  result.append(", ");
               }

               result.append(indexBufName);
               result.append(".");
               result.append(field);
               result.append(" ");
               result.append(nextComp.isDescending() ? "desc" : "asc");
            }
         }

         return result.toString();
      }
      
      /**
       * Compares the order of tables (only important in a multi-table query)
       * with the order of the sort criteria.  A jumble sort occurs (is
       * <code>true</code>) if the fields from two or more tables are
       * intermixed (fields from one table are processed both before AND 
       * after a field or fields of another table) AND/OR are out of order
       * (the fields of a later occurring table are less granular sort
       * criteria than fields of an earlier occurring table).  If there is
       * only 1 table or if there are no sort criteria, jumble sorting will
       * be <code>false</code>.
       *
       * @param    buffers
       *           A list of the tables by node ID, in the order of
       *           occurrence.
       * @param    sortCriteria
       *           A list of the tables referenced by each sort criteria,
       *           in the order of occurrence.
       *
       * @return   <code>true</code> if jumble sorting is detected.
       */
      public boolean detectJumbleSort(LinkedList buffers, 
                                      LinkedList sortCriteria)
      {
         boolean result = false;
         
         // quick out if this is not a multi-table query OR if there are no
         // BY clauses
         if (buffers.size() > 1 && sortCriteria.size() > 0)
         {         
            Set      processed = new HashSet();
            Iterator bufs      = buffers.iterator();
            Iterator bys       = sortCriteria.iterator();
            Long     bufid     = (Long) bufs.next();
            
            outer:
            while (bys.hasNext())
            {
               Long byid = (Long) bys.next();
               
               while (!byid.equals(bufid))
               {
                  if (processed.contains(byid))
                  {
                     // this is a reference to a previous table, so this is
                     // either out of order or intermix
                     result = true;
                     break outer;
                  }
                  
                  if (bufs.hasNext())
                  {
                     processed.add(bufid);
                     bufid = (Long) bufs.next();
                  }
                  else
                  {
                     // we are trying to reference a table that is 
                     // different than the last table, this must be an
                     // intermix
                     result = true;
                     break outer;
                  }                     
               }
            }
         }
            
         return result;
      }
      
      /**
       * Return the given named index for the current table.
       *
       * @param    name
       *           The name of the index to obtain.
       *
       * @return   The named index or <code>null</code> if there are no
       *           indexes explicitly defined for this table or if there are
       *           no matches.
       */
      public P2JIndex getIndex(String name)
      {
         Iterator iter = indexes();
         
         while (iter.hasNext())
         {
            P2JIndex result = (P2JIndex) iter.next();
            
            if (result.getName().equalsIgnoreCase(name))
            {
               // we are done
               return result;
            }
         }
         
         return null;
      }

      /**
       * Returns the name of the index which corresponds the given prefix.
       * If an index name completely matches the given prefix, it is returned.
       * If there is a single index which starts with the given prefix, it is
       * returned. If there are no indexes or more than one index that start
       * with the given prefix, <code>null</code> is returned.
       *
       * @param   prefix
       *          Prefix of the index name.
       *  
       * @return  The name of the index which corresponds the given prefix or
       *          <code>null</code> if there is no corresponding index. 
       */
      public String getIndexNameByPrefix(String prefix)
      {
         Iterator iter = indexes();
         String targetName = null;
         prefix = prefix.toLowerCase();

         while (iter.hasNext())
         {
            P2JIndex result  = (P2JIndex) iter.next();
            String indexName = result.getName().toLowerCase();

            if (indexName.startsWith(prefix))
            {
               if (indexName.length() == prefix.length())
               {
                  // we have complete match
                  return indexName;
               }

               if (targetName == null)
               {
                  targetName = indexName;
               }
               else
               {
                  // we have two or more matching indexes
                  return null;
               }
            }
         }

         return targetName;
      }
      
      /**
       * Return the index set as primary for the current table.
       *
       * @return   The primary index or <code>null</code> if there are no
       *           indexes explicitly defined for this table (as long as
       *           there is at least 1 explicit index, there is always
       *           a primary index).
       */
      public P2JIndex getPrimaryIndex()
      {
         Iterator iter = indexes();
         
         while (iter.hasNext())
         {
            P2JIndex result = (P2JIndex) iter.next();
            
            if (result.isPrimary())
            {
               // we are done
               return result;
            }
         }
         
         return null;
      }
      
      /**
       * Tests the given field name to determine if it is the last component
       * of an index and is flagged for abbreviation support.
       *
       * @param    name
       *           The field name.
       *
       * @return   <code>true</code> if this field is the last index
       *           component of an index AND is flagged for abbreviation
       *           support.
       */
      public boolean isAbbreviatedField(String name)
      {
         Iterator iter = indexes();
         P2JIndex next = null;
         
         while (iter.hasNext())
         {
            next = (P2JIndex) iter.next();
            Iterator comps = next.components();
            
            P2JIndexComponent nextComp = null;
            
            while (comps.hasNext())
            {
               nextComp = (P2JIndexComponent) comps.next();
               
               if (nextComp.getName().equalsIgnoreCase(name))
               {
                  if (nextComp.isAbbreviated())
                  {
                     return true;
                  }
               }
            }
         }
         
         return false;
      }
      
      /**
       * Add an equality match (for the given index component) to the list of
       * matches.
       *
       * @param    field
       *           The index component (field) name which was matched.
       */
      public void addEqualityMatch(String field)
      {
         equalityMatches.add(field);
      }
      
      /**
       * Add an begins match (for the given index component) to the list of
       * matches.
       *
       * @param    field
       *           The index component (field) name which was matched.
       */
      public void addBeginsMatch(String field)
      {
         beginsMatches.add(field);
      }
      
      /**
       * Add an range match (for the given index component) to the list of
       * matches.
       *
       * @param    field
       *           The index component (field) name which was matched.
       */
      public void addRangeMatch(String field)
      {
         rangeMatches.add(field);
      }
      
      /**
       * Add an sort match (for the given index component) to the list of
       * matches (and maintains the order of specificity based on the
       * order of calls to this method).
       *
       * @param    field
       *           The index component (field) name which was matched.
       * @param    descending
       *           <code>true</code> if the sort is a descending sort,
       *           <code>false</code> for an ascending sort.
       */
      public void addSortMatch(String field, Boolean descending)
      {
         sortMatches.put(field, descending);
      }
      
      /**
       * Add an word match (for the given index component) to the list of
       * matches.
       *
       * @param    field
       *           The index component (field) name which was matched.
       */
      public void addWordMatch(String field)
      {
         wordMatches.add(field);
      }
      
      /**
       * Creates a list of all possible indexes for the current table.
       *
       * @return   The list of all index objects (may be empty if no indexes
       *           exist for this table).
       */
      public List getFullIndexList()
      {
         List     all  = new ArrayList();
         Iterator iter = indexes();
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            // remember the complete list in case we find no matches
            all.add(idx);
         }
         
         return all;
      }
      
      /**
       * Uses the stored equality match data and the list of all indexes for
       * the current record to determine if there is a unique index that would
       * be matched (the index must be unique and all index components must
       * have equality matches).  Only the indexes with the most index 
       * components are returned.  For example, if 2 unique indexes are
       * selectable, the one with more index components will be returned.
       * In the case that more than one unique index with the same number of
       * components is selectable, multiple matches are disambiguated using
       * the {@link #breakTie} method (in descending order).
       *
       * @return   The unique index name selected or <code>null</code> if no
       *           unique indexes are selectable by this criteria).
       */
      private String findUniqueIndexMatch()
      {
         List     matches = new ArrayList();
         Iterator iter    = indexes();
         int      max     = 0;
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            if (idx.isUnique() && allComponentsMatch(idx, equalityMatches))
            {
               int num = idx.size();
               
               // only add this index if it has at least as many matches as
               // the previous candidates
               if (num >= max)
               {
                  // if this is the best match so far, reset the list 
                  if (num > max)
                  {
                     max = num;
                     matches.clear();
                  }
                  
                  // add the match
                  matches.add(idx);
               }
            }
         }
         
         return breakTie(matches, false);
      }
      
      /**
       * Uses the stored word match data and the list of indexes for the
       * current record to determine if there is a word index that would
       * be matched.  In the case that more than one index is selectable
       * the {@link #breakTie} method (in ascending order) is used to select
       * a single result.
       *
       * @return   The index name that was matched or <code>null</code> if
       *           no indexes were selectable by this criteria.
       */
      private String findWordIndexMatch()
      {
         List     matches = new ArrayList();
         Iterator iter    = indexes();

         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            if (idx.isWord() && numComponentsMatch(idx, wordMatches) > 0)
            {
               matches.add(idx);
            }
         }
         
         return breakTie(matches, true);
      }
      
      /**
       * Uses the equality and range match data and the list of selectable
       * indexes for the current record to determine if there is an index
       * that would be matched based on the largest number of sequential 
       * leading index components.  More than one index may remain selectable
       * after this method is complete.
       * <p>
       * Only indexes that have sequential leading index components matching
       * can be selectable by this method.  The method first checks for
       * equality matches and if multiple equivalent equality match choices
       * exist (multiple choices with the same number of leading sequential
       * equality matches), then a single subsequent field is checked for a
       * range match.  This provides a very specific/tightly coupled
       * tie break process.
       * <p>
       * The selection process removes indexes from the selectable list only
       * in the case where matches occur.  This means that those selectable
       * indexes will remain in the list and thus the returned list cannot
       * be empty unless there are no indexes at all for a given table.
       *
       * @param    selectable
       *           The list of index candidates for the current table. This
       *           may be equivalent to the full list of all possible indexes
       *           or it may be a subset if prior processing removed indexes
       *           form the selection list.
       *
       * @return   The list of remaining candidates that are selectable.
       */
      private List findLeadingIndexMatch(List selectable)
      {
         List     eMatches = new ArrayList();
         List     bMatches = new ArrayList();
         List     rMatches = new ArrayList();
         Iterator iter     = selectable.iterator();
         int      max      = 0;
         
         // check equality first
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            int num = leadingComponentsMatch(idx, 0, equalityMatches);
            
            // is this a candidate?
            if (num > 0)
            {
               // only add this index if it has at least as many matches as
               // the previous candidates
               if (num >= max)
               {
                  // if this is the best match so far, reset the list 
                  if (num > max)
                  {
                     max = num;
                     eMatches.clear();
                  }
                  
                  // add the match
                  eMatches.add(idx);
               }
            }
         }
         
         // if there is more than 1 match, check if there is a begins match 
         // on the next field
         if (eMatches.size() > 1)
         {
            // bypass range matching down-stream
            rangeBypass = true;
            
            iter = eMatches.iterator();
            
            while (iter.hasNext())
            {
               P2JIndex idx = (P2JIndex) iter.next();
               
               // only add a match if there are more fields than already
               // found AND the next field has a range match
               if (idx.size() > max &&
                   beginsMatches.contains(idx.getComponent(max).getName()))
               {
                  // add the match
                  bMatches.add(idx);
               }
            }
            
            // did we cut the list down further?
            if (bMatches.size() > 0)
            {
               return bMatches;
            }
         }
         
         // the begins match check did not cut down the list further, if 
         // there is more than 1 match, check if there is a range match on
         // the next field
         if (eMatches.size() > 1)
         {
            // bypass range matching down-stream
            rangeBypass = true;
            
            iter = eMatches.iterator();
            
            while (iter.hasNext())
            {
               P2JIndex idx = (P2JIndex) iter.next();
               
               // only add a match if there are more fields than already
               // found AND the next field has a range match
               if (idx.size() > max &&
                   rangeMatches.contains(idx.getComponent(max).getName()))
               {
                  // add the match
                  rMatches.add(idx);
               }
            }
            
            // did we cut the list down further?
            if (rMatches.size() > 0)
            {
               return rMatches;
            }
         }
         
         return (eMatches.size() > 0 ? eMatches : selectable);
      }
      
      /**
       * Uses the "begins" match data and the list of selectable indexes for
       * the current record to determine if there is an index that would
       * be matched.  More than one index may remain selectable after this
       * method is complete.
       * <p>
       * Only indexes that have the first (leading) index component matching
       * can be selectable.  All other begins matches are ignored.  There is
       * no concept of the "most" begins matches.
       * <p>
       * The selection process removes indexes from the selectable list only
       * in the case where matches occur.  This means that those selectable
       * indexes will remain in the list and thus the returned list cannot
       * be empty unless there are no indexes at all for a given table.
       *
       * @param    selectable
       *           The list of index candidates for the current table. This
       *           may be equivalent to the full list of all possible indexes
       *           or it may be a subset if prior processing removed indexes
       *           form the selection list.
       *
       * @return   The list of remaining candidates that are selectable.
       */
      private List findBeginsIndexMatch(List selectable)
      {
         List     matches = new ArrayList();
         Iterator iter    = selectable.iterator();
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            String field = idx.getComponent(0).getName();
            
            // if the first field is not an equality match but IS a begins
            // match, then it is a candidate
            if (beginsMatches.contains(field))
            {
               // add the match
               matches.add(idx);
            }
         }
         
         return (matches.size() > 0 ? matches : selectable);
      }
      
      /**
       * Uses the range match data and the list of selectable indexes for the
       * current record to determine if there is an index that would
       * be matched.  More than one index may remain selectable after this
       * method is complete.
       * <p>
       * Only indexes that have the first (leading) index component matching
       * can be selectable.  All other range matches are ignored.  There is
       * no concept of the "most" range matches.
       * <p>
       * The selection process removes indexes from the selectable list only
       * in the case where matches occur.  This means that those selectable
       * indexes will remain in the list and thus the returned list cannot
       * be empty unless there are no indexes at all for a given table.
       *
       * @param    selectable
       *           The list of index candidates for the current table. This
       *           may be equivalent to the full list of all possible indexes
       *           or it may be a subset if prior processing removed indexes
       *           form the selection list.
       *
       * @return   The list of remaining candidates that are selectable.
       */
      private List findRangeIndexMatch(List selectable)
      {
         List     matches = new ArrayList();
         Iterator iter    = selectable.iterator();
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            String field = idx.getComponent(0).getName();
            
            // if the first field is not an equality match but IS a range 
            // match, then it is a candidate
            if (rangeMatches.contains(field))
            {
               // add the match
               matches.add(idx);
            }
         }
         
         return (matches.size() > 0 ? matches : selectable);
      }
      
      /**
       * Uses the sort match data and the list of selectable indexes for the
       * current record to determine if there is an index that would
       * be matched.  In the case that more than one index is selectable
       * the {@link #breakTie} method (in ascending order) is used to select
       * a single result.
       * <p>
       * Only indexes that have sequential leading index components matching
       * can be selectable.  Sort matches must also agree on the direction
       * of the sort for the given index component (ascending/descending).
       * <p>
       * The selection process removes indexes from the selectable list only
       * in the case where matches occur.  This means that those selectable
       * indexes will remain in the list and thus the returned list cannot
       * be empty unless there are no indexes at all for a given table.
       *
       * @param    selectable
       *           The list of index candidates for the current table. This
       *           may be equivalent to the full list of all possible indexes
       *           or it may be a subset if prior processing removed indexes
       *           form the selection list.
       *
       * @return   The index name that was matched or <code>null</code> if
       *           no indexes were selectable by this criteria.
       */
      private String findSortIndexMatch(List selectable)
      {
         List     matches = new ArrayList();
         Iterator iter    = selectable.iterator();
         int      max     = 0;
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next();
            
            int num = leadingSortComponentsMatch(idx);
            
            // is this a candidate?
            if (num != 0)
            {
               // only add this index if it has at least as many matches as
               // the previous candidates
               if (Math.abs(num) >= Math.abs(max))
               {
                  // if this is the best match so far, reset the list 
                  if (Math.abs(num) > Math.abs(max))
                  {
                     max = num;
                     matches.clear();
                  }
                  
                  // add the match
                  matches.add(idx);
               }
            }
         }
         
         return breakTie((matches.size() > 0 ? matches : selectable), true);
      }
      
      /**
       * Calculates if all index components of the given index are found in
       * the match set.
       *
       * @param    idx
       *           The index to test.
       * @param    matches
       *           The list of matches that were found.
       *
       * @return   <code>true</code> if all index components were matched.
       */
      private boolean allComponentsMatch(P2JIndex idx, Set matches)
      {
         return (idx.size() == numComponentsMatch(idx, matches));
      }
      
      /**
       * Calculates how many index components of the given index are to be
       * found in the match set.
       *
       * @param    idx
       *           The index to test.
       * @param    matches
       *           The list of matches that were found.
       *
       * @return   The number of matches or 0 if there are no matches.
       */
      private int numComponentsMatch(P2JIndex idx, Set matches)
      {
         Iterator iter = idx.components();
         int      num  = 0;
         
         while (iter.hasNext())
         {
            P2JIndexComponent comp = (P2JIndexComponent) iter.next();
            
            if (matches.contains(comp.getName()))
            {
               num++;
            }
         }
         
         return num;
      }      
      
      /**
       * Calculates the number of sequential leading index components of the
       * given index are found in the match set starting at the specified
       * position in the index component list.
       *
       * @param    idx
       *           The index to test.
       * @param    start
       *           The 0-based position of the first component to compare.
       * @param    matches
       *           The list of matches that were found.
       *
       * @return   The number of sequential leading index components that
       *           were matched (may be 0 if no leading components were
       *           matched) starting at the given position.
       */
      private int leadingComponentsMatch(P2JIndex idx, int start, Set matches)
      {
         int len = idx.size();
         int num = 0;
         
         for (int i = start; i < len; i++)
         {
            P2JIndexComponent comp = idx.getComponent(i);
            
            if (matches.contains(comp.getName()))
            {
               // this a matched component
               num++;
            }
            else
            {
               // the first missing component stops the search
               break;
            }
         }
         
         return num;
      }
      
      /**
       * Calculates the number of sequential leading index components of the
       * given index are found in the sort match set (tests component name, 
       * the specificity of the sort matches and the sort direction for
       * compatibility). The set of components can be matched into direct or
       * reverse way, e.g. reverse match for the "crit1 asc crit2 desc crit3
       * desc" is the "crit1 desc crit2 asc crit3 asc". 
       *
       * @param    idx
       *           The index to test.
       *
       * @return   The absolute value of the returned value is the number of
       *           sequential leading index components that were matched. If
       *           the returned value is greater than <code>0</code> then
       *           components were matched in the direct way. If the returned
       *           value is less than <code>0</code> then components were
       *           matched in the reverse way. If the returned value is 
       *           <code>0</code> then no leading components were matched.
       */
      private int leadingSortComponentsMatch(P2JIndex idx)
      {
         Iterator iter = idx.components();
         Iterator sort = sortMatches.keySet().iterator();
         int      num  = 0;
         boolean  tryReverseMatch = false;
         int      iterNum = 0;
         
         while (iter.hasNext())
         {
            iterNum++;
            P2JIndexComponent comp = (P2JIndexComponent) iter.next();
            
            if (sort.hasNext())
            {
               String critName = (String) sort.next();
               
               if (critName.equals(comp.getName()))
               {
                  Boolean descend = (Boolean) sortMatches.get(critName);

                  if (iterNum == 1 && descend != comp.isDescending())
                  {
                     tryReverseMatch = true;
                  }
                  
                  if (tryReverseMatch ? descend != comp.isDescending() :
                                        descend == comp.isDescending())
                  {
                     // this a matched component
                     num++;
                  }
                  else
                  {
                     // this sort match is a different direction than the
                     // component of the index
                     break;
                  }                  
               }
               else
               {
                  // further sort matches are not from this index or are not
                  // sequential leading fields of this index
                  break;
               }               
            }
            else
            {
               // no more sort matches to check
               break;
            }            
         }
         
         return tryReverseMatch ? -num : num;
      }
      
      /**
       * If the given list has only a single entry, read and return the
       * index name, otherwise return <code>null</code>.
       *
       * @param    matches
       *           The list of <code>P2JIndex</code> objects to read.
       *
       * @return   The selected index name or <code>null</code> if no indexes
       *           or more than 1 index are in the list.
       */
      private String readSingleIndex(List matches)
      {
         if (matches.size() != 1)
         {
            return null;
         }
         
         P2JIndex idx = (P2JIndex) matches.get(0);
         
         return idx.getName();
      }
         
      /**
       * Select one index from the given list of indexes that are
       * selectable, if there are more than 1 in the list.  If there is only
       * 1 index in the list, this is returned.  The precedence for the
       * choice is if one of the indexes is the primary index, this is 
       * returned, otherwise the first or last index name alphabetically is
       * returned (the ascending or descending order is based on the
       * <code>ascend</code> parameter).
       *
       * @param    matches
       *           The list of <code>P2JIndex</code> objects to select from.
       * @param    ascend
       *           <code>true</code> to compare lexicographically (ascending  
       *           order) and <code>false</code> to compare in reverse 
       *           lexicographical order (descending order).
       *
       * @return   The selected index or <code>null</code> if no indexes
       *           are in the list.
       */
      private String breakTie(List matches, boolean ascend)
      {
         String result  = null;
         String alpha   = null;
         String upper   = null; 
         int    compare = 0;
         
         Iterator iter = matches.iterator();
         
         while (iter.hasNext())
         {
            P2JIndex idx = (P2JIndex) iter.next(); 
            
            if (idx.isPrimary())
            {
               return idx.getName();
            }
            
            alpha = idx.getName();
            upper = alpha.toUpperCase();
            
            if (result != null)
            {
               compare = result.toUpperCase().compareTo(upper);
            }
            
            // init if first time through, otherwise compare lexicographically
            // with the next name (we uppercase to get the same kind of
            // case-insensitive comparison that Progress does)
            if (result == null || (ascend ? compare > 0 : compare < 0))
            {
               result = alpha;
            }
         }
         
         return result;
      }
   }   
}
