/*
** Module   : DenormalizationMapper.java
** Abstract : This class keeps information about properties for denormalized fields and mapping
*             historical methods to custom methods.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 VMN 20140129 Created initial version.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.Method;
import java.util.*;

/**
 * This class keeps information about properties for denormalized fields and mapping
 * historical methods to custom methods.
 */
public class DenormalizationMapper
{
   /** Map of historical methods to custom methods for denormalized fields with extent.*/
   private final Map<String, List<Method>> method2CustomMethods;

   /** Collection of properties for denormalized fields with extent.*/
   private final Collection<String> denormalizedProperties;

   /**
    * Constructor.
    *
    * @param    dmoImplementationClass
    *           DMO implementation class.
    */
   public DenormalizationMapper(Class<?> dmoImplementationClass)
   {
      method2CustomMethods = new HashMap<>();
      denormalizedProperties = new HashSet<>();
      Method[] methods = dmoImplementationClass.getDeclaredMethods();
      try {
         for(Method method : methods)
         {
            String methodName = method.getName();
            Class<?>[] args = method.getParameterTypes();
            CustomExtentMethods customExtentMethods =
                  method.getAnnotation(CustomExtentMethods.class);
            if (customExtentMethods != null)
            {
               String property = method.getAnnotation(CustomExtentProperty.class).name();
               denormalizedProperties.add(property);

               CustomExtentMethod[] methodNames = customExtentMethods.value();
               List<Method> customMethods = new ArrayList<>();
               for (CustomExtentMethod customMethodName : methodNames)
               {
                  Class<?>[] arg = args.length == 1 ? new Class<?>[0] : new Class<?>[]{args[1]};
                  Method customMethod =
                        dmoImplementationClass.getDeclaredMethod(customMethodName.name(), arg);
                  customMethods.add(customMethod);
               }
               method2CustomMethods.put(methodName, customMethods);
            }
         }
      }
      catch (NoSuchMethodException e)
      {
         e.printStackTrace();
      }
   }

   /**
    * Report whether given property corresponds to denormalized field with extent.
    *
    * @param    propertyName
    *           Property name.
    *
    * @return  <code>true</code> if property corresponds to denormalized field with extent;
    * otherwise <code>false</code>.
    */
   public boolean isCustomExtentProperty(String propertyName)
   {
      return denormalizedProperties.contains(propertyName);
   }

   /**
    * Returns custom getter or setter method for denormalized field for given indexed
    * getter/setter method and index.
    *
    * @param    method
    *           Method.
    * @param    getter
    *           If <code>true</code> the resulting method will be a getter.
    * @param    index
    *           Zero-based index of custom property name for denormalized field with extent.
    *
    * @return  Custom getter or setter method for denormalized field with extent or
    *          <code>null</code> if such method does not exist.
    */
   public Method getCustomMethod(Method method, boolean getter, int index)
   throws PersistenceException
   {
      String methodName = method.getName();
      if (getter)
      {
         methodName = "get" + methodName.substring(3);
      }
      if (!method2CustomMethods.containsKey(methodName))
      {
         return null;
      }
      List<Method> customMethods = method2CustomMethods.get(methodName);

      if (index < 0 || index >= customMethods.size())
      {
         throw new PersistenceException(
            "Custom method for extent field with index " + index + " not found");
      }
      else
      {
         return customMethods.get(index);
      }
   }
}
