/*
** Module   : P2OLookup.java
** Abstract : Helper class which loads P2O ASTs and enables lookup access
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 ECF 20051013   @23034 Created initial version. Lazily loads P2O
**                           ASTs and enables lookups of Java names for
**                           legacy Progress table and field names.
**                           This first implementation supports fixed
**                           tables only (not temp- or work-tables).
**                           Also exposes index information by table.
** 002 ECF 20051017   @23054 Change to P2JIndex construction parameters.
**                           Use historical index name instead of
**                           converted name.
** 003 ECF 20051023   @23095 Added javaPropertyName() convenience method.
**                           Looks up a Java property name for a given
**                           legacy field name, optionally qualifying it
**                           with a buffer name.
** 004 ECF 20051027   @23176 Added 'ignore-case' processing for index
**                           components. Takes new annotation in P2O AST
**                           into account when creating indexes.
** 005 ECF 20051102   @23181 Added legacyFieldNames method. Retrieves set
**                           of legacy field names for a Progress table,
**                           in display order.
** 006 ECF 20051121   @23423 Added isValidatable method. Indicates whether
**                           a schema field must invoke validation
**                           processing before it may be saved to its DMO.
** 007 ECF 20060202   @24252 Added support for temp/work-table schema
**                           data. These data are refreshed by pattern
**                           workers each time a new source file AST is
**                           visited.
** 008 ECF 20060227   @24818 Minor fix for field-level validation support.
**                           Made check for validatable annotation safer.
** 009 ECF 20060408   @25429 Added isDMOInterface() method. Indicates
**                           whether a given class name collides with any
**                           known DMO interface name. Temp table DMOs not
**                           currently handled.
** 010 ECF 20061026   @30795 Added isForeignReference() method. Indicates
**                           whether a given table is referenced by the
**                           foreign key of any other table.
** 011 ECF 20061101   @30882 Added dumpTableDependencies() method. Debug
**                           method to print tree of dependencies for a
**                           table.
** 012 ECF 20070628   @35342 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 013 ECF 20080304   @37468 Modified to support changes in P2JIndex.
** 014 SVL 20080326   @37683 Synthetic indexes are ignored in the
**                           initializeIndexes().
** 015 GES 20090424   @41953 Import change.
** 016 GES 20090429   @42051 Match package and class name changes.
** 017 GES 20090515   @42220 Moved to AstManager from AstRegistry/AstPersister.
** 018 GES 20090518   @42398 Import change.
** 019 LMR 20110106          Fixes to reflect renaming of load() method in 
**                           Configuration class to getSchemaConfig() and 
**                           removal of its 'key' parameter.
** 020 ECF 20130130          Add metadata to P2OLookup data structures, under each primary schema
**                           loaded. Implemented generics.
** 021 OM  20130305          Fix java path for metadata tables. Added message when a table name 
**                           conflict is detected.
** 022 ECF 20131013          Don't add synthetic legacy fields to legacy field name map.
** 023 CA  20131028          All the static info is kept context-local, for runtime query
**                           conversion support.  All filenames must be normalized.
** 024 OM  20131030          Added support for checking if a schema contains a certain dmo iface.
** 025 VMN 20140129          Added support for denormalized properties.
*/

package com.goldencode.p2j.schema;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;

/**
 * A helper class which exposes information stored in P2O ASTs for access
 * during source conversion processing. Information is exposed via static
 * methods, which resolve lookups using individual instances of this class.
 * The following data currently is exposed:
 * <ul>
 *   <li>Java DMO class names (given legacy table names)
 *   <li>Java DMO property names (given legacy field names)
 *   <li>Java DMO package names (given legacy table names)
 *   <li>Index descriptors (given legacy table names)
 *   <li>Whether a field has schema-level validation configured (given legacy
 *       field names)
 *   <li>Whether a DMO interface exists in any schema (given a Java class
 *       name)
 *   <li>Whether a database table is referenced by the foreign key of any
 *       other table (given the legacy name of the table which is the
 *       potential referent)
 * </ul>
 *
 * @see  P2JIndex
 * @see  P2JIndexComponent
 */
public final class P2OLookup
implements DataModelTokenTypes
{
   /** Name of schema dictionary configuration */
   public static final String CFG_NAME = "schema-dict";
   
   /** File extension appended to schema and source filenames */
   public static final String P2O_POSTFIX = ".p2o";
   
   /** Key for "package" database AST annotation */
   public static final String PACKAGE = "package";
   
   /** Key for "historical" class/property AST annotation */
   public static final String HISTORICAL = "historical";

   /** Key for "datatype" property AST annotation */
   public static final String DATATYPE = "datatype";

   /** Key for "extent" property AST annotation */
   public static final String EXTENT = "extent";
   
   /** Key for "validatable" property AST annotation */
   public static final String VALIDATABLE = "validatable";
   
   /** Key for "table" class AST annotation */
   public static final String TABLE = "table";
   
   /** Key for "unique" index AST annotation */
   public static final String UNIQUE = "unique";
   
   /** Key for "primary" index AST annotation */
   public static final String PRIMARY = "primary";
   
   /** Key for "synthetic" index AST annotation */
   public static final String SYNTHETIC = "synthetic";
   
   /** Key for "word" index AST annotation */
   public static final String WORD = "word";
   
   /** Key for "refid" index column AST annotation */
   public static final String REFID = "refid";
   
   /** Key for "descend" index column AST annotation */
   public static final String DESCEND = "descend";
   
   /** Key for "ignore-case" index column AST annotation */
   public static final String IGNORE_CASE = "ignore-case";
   
   /** Key for "abbreviated" index column AST annotation */
   public static final String ABBREVIATED = "abbreviated";
   
   /** Key for "foreign" relation AST annotation */
   public static final String FOREIGN = "foreign";
   
   /** Key for "references" relation AST annotation */
   public static final String REFERENCES = "references";
   
   /** Key for current temp table schema in schema map */
   private static final String TEMP_TABLE_KEY = null;
   
   /** Container with user-specific data. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /** Stores the list of meta tables */
   private final Set<String> historicalMetaTables = new HashSet<String>();

   /** Map of AST IDs to class ASTs */
   private final Map<Long, Aast> classByIDMap = new HashMap<Long, Aast>();
   
   /** Map of historical table/field names to Java class/property names */
   private final Map<String, String> javaNameMap = new LinkedHashMap<String, String>();

   /** Set of historical fully qualified denormalized properties of fields with custom extent */
   private final Set<String> denormalizedProperties = new HashSet<>();

   /** Map of historical table names to index descriptors */
   private final Map<String, List<P2JIndex>> indexMap = new HashMap<String, List<P2JIndex>>();
   
   /** Map of historical table names to sets of referencing tables */
   private final Map<String, Set<Long>> foreignReferenceMap = new HashMap<String, Set<Long>>();
   
   /** Map of property AST unique IDs to the AST nodes. */
   private final Map<Long, Aast> propsMap = new HashMap<Long, Aast>();
   
   /** Set of schema names of validatable fields */
   private final Set<String> validatables = new HashSet<String>();
   
   /** Set of all unqualified DMO Java interface names */
   private final Set<String> javaNames = new HashSet<String>();
   
   /** Java package associated with this lookup instance */
   private String javaPackage = null;
   
   /** Java package associated with this lookup instance metadata */
   private String metaPackage = null; 
   
   /**
    * Constructor which initializes this object with the information in the
    * P2O AST associated with <code>schema</code>.
    *
    * @param   schema
    *          Database schema name which determines which P2O AST to load.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   private P2OLookup(String schema)
   throws SchemaException
   {
      initialize(schema);
      propsMap.clear();
      locate().schemaMap.put(schema, this);
   }
   
   /**
    * Constructor which initializes this object with the information in the
    * P2O AST stored in <code>filename</code>.  This variant is intended to
    * be used to load temp/work-table schema information.  If there is no
    * file with name <code>filename</code>, we still update the static
    * schema map, in order to replace the <code>P2OLookup</code> instance
    * (if any) already associated with <code>schema</code>.  This ensures
    * that we are not using out-of-date temp table schema information.
    *
    * @param   schema
    *          Database schema name to which this instance is mapped.
    * @param   filename
    *          Name of file containing the P2O AST to load.  File may or may
    *          not exist.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   private P2OLookup(String schema, String filename)
   throws SchemaException
   {
      if (AstManager.get().isExistingAst(filename))
      {
         try
         {
            initialize(schema, AstManager.get().loadTree(filename), false);
         }
         catch (AstException exc)
         {
            throw new SchemaException(
               "Error loading P2O AST for database schema:  " + schema,
               exc);
         }
         propsMap.clear();
      }
      
      locate().schemaMap.put(schema, this);
   }
   
   /**
    * Attempt to load the P2O AST stored in <code>filename</code>, which must
    * represent a temp/work-table schema.  This method should be invoked by
    * conversion service pattern workers which need to load temporary schema
    * information upon loading a new source file AST.  A new instance of this
    * class is created as a result, and is stored in the schema map, under
    * the temp table schema key.  If this method is invoked more than once in
    * succession with the same file name, only the first invocation triggers
    * the P2O AST to be loaded.
    * <p>
    * If no file exists with the name <code>filename</code>, an empty
    * instance of this class is stored in the schema map under the temp table
    * schema key.  This ensures that an out-of-date <code>P2OLookup</code>
    * instance is not maintained under the temp table key beyond it useful
    * life.
    *
    * @param   filename
    *          Name of file containing the P2O AST to load.  File may or may
    *          not exist, but this value may not be <code>null</code>.
    *
    * @throws  RuntimeException
    *          if there is any error loading or parsing the P2O AST data for
    *          the given file;  non-existence of the file is not considered
    *          an error.
    */
   public static void loadTemporarySchema(String filename)
   {
      filename += P2O_POSTFIX;
      filename = Configuration.normalizeFilename(filename);

      if (!filename.equals(locate().lastTempFilename))
      {
         try
         {
            new P2OLookup(TEMP_TABLE_KEY, filename);
         }
         catch (SchemaException exc)
         {
            throw new RuntimeException(exc);
         }
         
         locate().lastTempFilename = filename;
      }
   }
   
   /**
    * Report whether the given class name is the unqualified name of any DMO
    * interface, in any schema tracked by the current P2J configuration.
    *
    * @param   name
    *          An unqualified class name.
    *
    * @return  <code>true</code> if <code>name</code> collides with any DMO
    *          interface in any schema, else <code>false</code>.
    *
    * @throws  SchemaException
    *          if there is an error loading schema names from configuration.
    */
   public static boolean isDMOInterface(String name)
   throws SchemaException
   {
      Iterator<String> iter = schemaNames();
      while (iter.hasNext())
      {
         String schema = iter.next();
         if (isDMOInterface(name, schema))
         {
            return true;
         }
      }
      
      return false;
   }

   /**
    * Report whether the given class name is the unqualified name of a DMO interface, in the
    * <code>schema<code/> tracked by the current P2J configuration.
    *
    * @param   name
    *          An unqualified class name.
    * @param   schema
    *          The name of teh schema to look into.
    *
    * @return  <code>true</code> if <code>name</code> is present as java name of a table from
    *          the specified <code>schema<code/> and <code>false<code/> if <code>schema<code/>
    *          does not exist or it does not contain such interface.
    *
    * @throws  SchemaException
    *          if there is an error loading schema names from configuration.
    */
   public static boolean isDMOInterface(String name, String schema)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(schema, EntityName.DATABASE);
      return lookup.containsDMOInterface(name);
   }
   
   /**
    * Report whether the given, historical table will be mapped in the
    * converted application to a database table which represents a foreign
    * reference of another table's foreign key.
    *
    * @param   historical
    *          Fully qualified, legacy Progress table name.
    *
    * @return  <code>true</code> if the given table represents the foreign
    *          end of a one-to-one or a one-to-many relation between two
    *          tables.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   public static boolean isForeignReference(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.TABLE);
      
      return lookup.foreignReferenceMap.containsKey(historical);
   }
   
   /**
    * Dump an exhaustive tree of nested table dependencies.  A dependency is
    * a foreign key relation from the dependent table to the "parent" table.
    * The tree is printed to stdout.  Each indent indicates a dependency to
    * the next outer level.
    *
    * @param   historical
    *          Fully qualified, legacy Progress table name.
    *
    * @throws  SchemaException
    *          if there is any error loading the AST from persistence.
    */
   public static void dumpTableDependencies(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.TABLE);
      lookup.dumpTableDependencies(historical, 0);
   }
   
   /**
    * Retrieve the Java package name of the DMO associated with the given
    * legacy Progress table name.
    *
    * @param   historical
    *          Fully qualified, legacy, Progress schema name for a table.
    *
    * @return  Corresponding Java package name, or <code>null</code> if no
    *          associated DMO or package name was found.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   public static String javaPackage(String historical)
   throws SchemaException
   {
      P2OLookup inst = getInstance(historical, EntityName.TABLE);
      return inst.getJavaPackage(inst.historicalMetaTables.contains(historical));
   }
   
   /**
    * Retrieve a Java class or variable name, based upon the legacy Progress
    * schema name of the corresponding table or field, respectively.
    *
    * @param   historical
    *          Fully qualified, legacy, Progress schema name for a table or
    *          field.
    * @param   type
    *          <code>EntityName</code> constant to indicate whether
    *          <code>historical</code> represents a table or a field.
    * @param   index
    *          Zero-based index of custom property name for denormalized field with extent,
    *          otherwise null.
    *
    * @return  Corresponding, unqualified, Java class or variable name.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence;
    *          if the table or field portion of <code>historical</code>
    *          cannot be resolved to a corresponding Java name.
    */
   public static String javaName(String historical, int type, Integer index)
   throws SchemaException
   {
      String name = getInstance(historical, type).getJavaName(historical, index);
      if (name == null)
      {
         throw new SchemaException(
            "No java name found for legacy name:  " + historical);
      }
      
      return name;
   }
   
   /**
    * Retrieve the full set of fully qualified, legacy field names for a
    * legacy Progress table.  The set is guaranteed to iterate in the default
    * display order defined for the original, Progress schema for that table.
    *
    * @param   historical
    *          Fully qualified, legacy, Progress schema name for a table.
    *
    * @return  Set of fully qualified, legacy field names.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence;
    *          if <code>historical</code> does not represent a known table.
    */
   public static Set<String> legacyFieldNames(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.TABLE);
      Set<String> names = lookup.getLegacyFieldNames(historical);
      if (names.isEmpty())
      {
         throw new SchemaException(
            "No table found for legacy name:  " + historical);
      }
      
      return names;
   }
   
   /**
    * Retrieve a Java property name, based upon the legacy Progress schema
    * name of the corresponding table or field, respectively.  If specified,
    * the property name will be qualified with a buffer name, suitable for
    * insertion into an HQL where clause or order by clause.  The convention
    * for the buffer name qualifier is to use the short DMO class name, with
    * the first letter lowercased.
    *
    * @param   historical
    *          Fully qualified, legacy, Progress schema name for a table or
    *          field.
    * @param   qualified
    *          <code>true</code> if property name should be qualified with a
    *          buffer name;  <code>false</code> if it should remain
    *          unqualified.
    * @param   index
    *          Zero-based index of custom property name for denormalized field with extent,
    *          otherwise null.
    *
    * @return  Corresponding Java property name.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence;
    *          if the table or field portion of <code>historical</code>
    *          cannot be resolved to a corresponding Java name.
    */
   public static String javaPropertyName(String historical, boolean qualified, Integer index)
   throws SchemaException
   {
      String property = javaName(historical, EntityName.FIELD, index);
      if (!qualified)
      {
         return property;
      }
      
      historical = historical.substring(0, historical.lastIndexOf("."));
      String className = javaName(historical, EntityName.TABLE, null);
      StringBuilder buf = new StringBuilder();
      buf.append(StringHelper.changeCase(className, false));
      buf.append('.');
      buf.append(property);
      
      return buf.toString();
   }
   
   /**
    * Get an iterator on all the index descriptors associated with the given,
    * legacy table. There is no guaranteed order of this iteration; it is
    * dictated by the order in which in which the indexes were read from the
    * exported Progress schema.
    *
    * @param   historical
    *          Fully qualified, legacy Progress table name.
    *
    * @return  Iterator on all indexes associated with the table named
    *          <code>historical</code>.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   public static Iterator<P2JIndex> indexes(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.TABLE);
      List<P2JIndex> indexList = lookup.getIndexes(historical);
      
      if (indexList == null)
      {
         throw new SchemaException(
            "Table '" + historical + "' not recognized");
      }
      
      return indexList.iterator();
   }
   
   /**
    * Compose a bean-like getter/setter method name from the Java DMO
    * property associated with the specified, legacy, Progress field name.
    * The name of the property is capitalized and prepended with
    * <code>prefix</code>.
    *
    * @param   prefix
    *          Verb to prepend to the capitalized property name; typically,
    *          "get", "is", or "set".
    * @param   historical
    *          Fully qualified, legacy, Progress schema name for a database
    *          field.
    * @param   index
    *          Zero-based index of custom property name for denormalized field with extent,
    *          otherwise null.
    *
    * @return  Method name as described above.
    *
    * @throws  SchemaException
    *          if this request results in the loading of a persisted P2O AST
    *          and there is any error loading P2J schema configuration data
    *          or loading the AST from persistence;  if the field portion of
    *          <code>historical</code> cannot be resolved to a corresponding
    *          Java name.
    */
   public static String beanMethodName(String prefix, String historical, Integer index)
   throws SchemaException
   {
      StringBuilder buf = new StringBuilder(prefix);
      String prop = javaName(historical, EntityName.FIELD, index);
      buf.append(StringHelper.changeCase(prop, true));
      
      return buf.toString();
   }

   /**
    * Check if given historical fully qualified property is denormalized.
    *
    * @param    historical
    *           Historical fully qualified property name.
    *
    * @return  <code>true</code> if given property is denormalized.
    */
   public static boolean isDenormalizedProperty(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.FIELD);

      return lookup.denormalizedProperties.contains(historical);
   }

   /**
    * Indicate whether a DMO property can be validated at runtime, given its
    * legacy schema name.  Possible forms of validation include:
    * <ul>
    *   <li>mandatory (non-nullable) constraint;
    *   <li>participation in a unique constraint;
    *   <li>explicit validation method (converted from a Progress validation
    *       expression).
    * </ul>
    *
    * @param   historical
    *          Fully qualified, legacy Progress field name.
    *
    * @return  <code>true</code> if the field should be validated at runtime,
    *          else <code>false</code>.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   public static boolean isValidatable(String historical)
   throws SchemaException
   {
      P2OLookup lookup = getInstance(historical, EntityName.FIELD);
      
      return lookup.validatables.contains(historical);
   }
   
   /**
    * Get an iterator on all schema names managed by the current P2J
    * configuration.
    *
    * @return  Iterator on all known schema names.
    *
    * @throws  SchemaException
    *          if there is an error loading schema names from configuration.
    */
   private static Iterator<String> schemaNames()
   throws SchemaException
   {
      if (locate().schemaNames == null)
      {
         locate().schemaNames = new HashSet<String>();
         SchemaConfig config = Configuration.getSchemaConfig();
         String meta = config.getMetadata().getName();
         Iterator<String> iter = config.databases();
         while (iter.hasNext())
         {
            String next = iter.next();
            if (!next.equals(meta))
            {
               locate().schemaNames.add(next);
            }
         }
      }
      
      return locate().schemaNames.iterator();
   }
   
   /**
    * Get the context-local {@link WorkArea} instance.
    * 
    * @return   See above.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /**
    * Get or create a <code>P2OLookup</code> instance, given a schema name or
    * a fully qualified, legacy table or field name. If the specified schema
    * has not been processed previously, a new instance is created and
    * cached;  otherwise, a cached instance is returned.
    *
    * @param   historical
    *          Schema name or fully qualified, legacy table or field name.
    * @param   type
    *          <code>EntityName</code> type, indicating whether the name
    *          <code>historical</code> represents a schema, table, or field.
    *
    * @return  A <code>P2OLookup</code> instance.
    */
   private static P2OLookup getInstance(String historical, int type)
   throws SchemaException
   {
      EntityName eName = new EntityName(type, historical);
      String schema = eName.getDatabase();
      P2OLookup lookup = locate().schemaMap.get(schema);
      if (lookup == null)
      {
         lookup = new P2OLookup(schema);
      }
      
      return lookup;
   }
   
   /**
    * Get the Java package name for the DMO interface classes associated with
    * this lookup object's schema.
    * 
    * @param   meta
    *          Return the metadata package name.
    *
    * @return  DMO interface package for this schema.
    */
   private String getJavaPackage(boolean meta)
   {
      return meta ? metaPackage : javaPackage;
   }

   /**
    * Report whether the given class name is the unqualified name of any DMO
    * interface in the schema represented by this lookup object.
    *
    * @param   name
    *          An unqualified class name.
    *
    * @return  <code>true</code> if <code>name</code> collides with any DMO
    *          interface in this schema, else <code>false</code>.
    */
   private boolean containsDMOInterface(String name)
   {
      return javaNames.contains(name);
   }
    
   /**
    * Get the converted, Java name of the class or property associated with
    * the given, legacy table or field name (respectively).
    *
    * @param   historical
    *          Fully qualified, legacy table or field name.
    * @param   index
    *          Zero-based index of custom property name for denormalized field with extent,
    *          otherwise null.
    *
    * @return  Unqualified, Java class name (for a table) or property name
    *          (for a field).
    */
   private String getJavaName(String historical, Integer index)
   {
      if (index == null)
      {
         if (historical.contains(","))
         {
            historical = historical.substring(0, historical.indexOf(','));
         }
      }
      else
      {
         historical = historical + "," + (index + 1);
      }
      return javaNameMap.get(historical);
   }
   
   /**
    * Get the set of legacy field names for a particular legacy table.
    *
    * @param   historical
    *          Fully qualified, legacy table or field name.
    *
    * @return  Set of legacy field names for the given table.
    */
   private Set<String> getLegacyFieldNames(String historical)
   {
      Set<String> names = new LinkedHashSet<String>();
      String test = historical + ".";
      Iterator<String> iter = javaNameMap.keySet().iterator();
      while (iter.hasNext())
      {
         String next = iter.next();
         if (next.startsWith(test))
         {
            names.add(next);
         }
      }
      
      return names;
   }
   
   /**
    * Get the full list of index descriptors associated with a Progress
    * table.
    *
    * @param   historical
    *          Fully qualified, legacy table name.
    *
    * @return  A list of index desriptors associated with the table named
    *          <code>historical</code>. The list may be empty, but will not
    *          be <code>null</code>.
    */
   private List<P2JIndex> getIndexes(String historical)
   {
      return indexMap.get(historical);
   }
   
   /**
    * Initialize this <code>P2OLookup</code> instance by loading the P2O AST associated with the
    * specified schema, walking it, and extracting and organizing the relevant information.
    * Currently, initialization does the following:
    * <ul>
    *   <li>maps fully qualified, legacy schema table and field names to their converted
    *       (unqualified) Java equivalents;
    *   <li>maps a table's fully qualified, legacy schema table name to a list of index
    *       descriptors defined for that table.
    * </ul>
    * <p>
    * After the primary schema is walked, we also walk the metadata schema, loading all of its
    * elements under the primary schema. This ensures all metadata references are resolved to the
    * correct database.
    *
    * @param   schema
    *          Name of database schema associated with the target P2O AST.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   private void initialize(String schema)
   throws SchemaException
   {
      Aast root = null;
      
      // load and initialize the primary schema and then the metadata schema, with the latter
      // being initialized under the former, so that references to metadata elements will match
      // the correct database
      String[] schemas = null;
      SchemaConfig.Metadata meta = Configuration.getSchemaConfig().getMetadata();
      if (!meta.getTables().isEmpty())
      {
         schemas = new String[] { schema, meta.getName() };
      }
      else
      {
         schemas = new String[] { schema };
      }
      
      for (String next : schemas)
      {
         try
         {
            root = loadAst(next);
         }
         catch (Exception exc)
         {
            throw new SchemaException("Error loading P2O AST for database schema:  " + next, exc);
         }
         
         initialize(schema, root, next.equals(meta.getName()));
      }
   }
   
   /**
    * Initialize this <code>P2OLookup</code> instance by walking the given
    * P2O AST and extracting and organizing the relevant information.
    * Currently, initialization does the following:
    * <ul>
    *   <li>maps fully qualified, legacy schema table and field names to
    *       their converted (unqualified) Java equivalents;
    *   <li>maps a table's fully qualified, legacy schema table name to a
    *       list of index descriptors defined for that table.
    * </ul>
    *
    * @param   schema
    *          Name of database schema associated with the target P2O AST.
    * @param   root
    *          Root node of P2O AST to inspect.
    * @param   meta
    *          Flags the schema as metadata information.
    *
    * @throws  SchemaException
    *          if there is any error loading P2J schema configuration data;
    *          if there is any error loading the AST from persistence.
    */
   private void initialize(String schema, Aast root, boolean meta)
   throws SchemaException
   {
      // Map the Java package to the database (i.e., schema) name.
      if (meta)
      {
         metaPackage = root.getAnnotation(PACKAGE) + "._meta";
      }
      else 
      {
         StringBuilder buf = new StringBuilder();
         buf.append(root.getAnnotation(PACKAGE))
            .append('.')
            .append(schema != null ? schema : SchemaDictionary.TEMP_TABLE_DB);
         javaPackage = buf.toString();
      }
      
      // Walk the class and property nodes, mapping each fully qualified,
      // historical schema name to an unqualified, converted Java name.
      // The Java names of classes are capitalized; those of properties
      // are not.
      Aast nextClass = root.getImmediateChild(CLASS, null);
      while (nextClass != null)
      {
         classByIDMap.put(nextClass.getId(), nextClass);
         
         String ifaceName = nextClass.getText();
         
         // Store the name in a set for quick tests later by isDMOInterface().
         if (!javaNames.add(ifaceName))
         {
            System.out.println("Duplicate dmo interface name detected: [" + ifaceName + "]");
         }
         
         StringBuilder buf = new StringBuilder();
         if (schema != null)
         {
            buf.append(schema).append(".");
         }
         buf.append(nextClass.getAnnotation(HISTORICAL));
         String histTable = buf.toString();
         javaNameMap.put(histTable, ifaceName);

         // store the meta tables, if the case
         if (meta)
         {
            historicalMetaTables.add(histTable);
         }

         // Retrieve first property in this class.
         Aast nextProp = nextClass.getImmediateChild(PROPERTY, null);
         while (nextProp != null)
         {
            // Compose fully qualified, legacy field name.
            buf = new StringBuilder(histTable);
            buf.append('.').append(nextProp.getAnnotation(HISTORICAL));
            String legacyName = buf.toString();
            
            // Map legacy field name to Java property name.
            if (!nextProp.isAnnotation(SYNTHETIC))
            {
               if (nextProp.isAnnotation("customextent"))
               {
                  String key = legacyName.substring(0, legacyName.indexOf(','));
                  if (!denormalizedProperties.contains(key))
                  {
                     String pseudoPropertyName =
                        (String) nextProp.getAnnotation("pseudoPropertyName");
                     denormalizedProperties.add(key);
                     javaNameMap.put(key, pseudoPropertyName);
                  }
               }
               javaNameMap.put(legacyName, nextProp.getText());
            }
            
            // Add to set of validatable properties if annotation is present.
            if (nextProp.isAnnotation(VALIDATABLE) &&
                ((Boolean) nextProp.getAnnotation(VALIDATABLE)).booleanValue())
            {
               validatables.add(legacyName);
            }
            
            // Store property AST by ID for reference later during
            // initialization.
            propsMap.put(nextProp.getId(), nextProp);
            
            // Retrieve next property in this class.
            nextProp = nextClass.getImmediateChild(PROPERTY, nextProp);
         }
         
         // Process indexes.
         initializeIndexes(histTable, nextClass);
         
         // Process relations.
         initializeRelations(histTable, nextClass);
         
         // Retrieve next class in this schema.
         nextClass = root.getImmediateChild(CLASS, nextClass);
      }
   }
   
   /**
    * Compose a list of index descriptors for the index children of the
    * given P2O class node and map it to the given, legacy table name.
    *
    * @param   historical
    *          Fully qualified, legacy table name.
    * @param   classAst
    *          P2O AST node describing a DMO; the direct parent of all P2O
    *          index nodes.
    */
   private void initializeIndexes(String historical, Aast classAst)
   {
      List<P2JIndex> indexList = new ArrayList<P2JIndex>();
      Aast nextIndex = classAst.getImmediateChild(INDEX, null);
      while (nextIndex != null)
      {
         if (!nextIndex.isAnnotation(SYNTHETIC))
         {
            P2JIndex indexDef = new P2JIndex(
                  (String) classAst.getAnnotation(TABLE),
                  (String) nextIndex.getAnnotation(HISTORICAL),
                  nextIndex.isAnnotation(UNIQUE),
                  nextIndex.isAnnotation(PRIMARY),
                  nextIndex.isAnnotation(WORD));
            indexList.add(indexDef);
            
            // Retrieve first component in the index.
            Aast nextCol = nextIndex.getImmediateChild(INDEX_COL, null);
            while (nextCol != null)
            {
               // Retrieve the property referenced by the index column.
               Long refid = (Long) nextCol.getAnnotation(REFID);
               Aast propAst = (Aast) propsMap.get(refid);
               
               // Add a component to the index descriptor.
               StringBuilder buf = new StringBuilder(historical);
               buf.append('.')
                  .append(propAst.getAnnotation(HISTORICAL));
               indexDef.addComponent(
                  buf.toString(),
                  nextCol.isAnnotation(DESCEND),
                  "character".equals(propAst.getAnnotation(DATATYPE)),
                  nextCol.isAnnotation(IGNORE_CASE),
                  nextCol.isAnnotation(ABBREVIATED));
               
               // Retrieve next component in the index.
               nextCol = nextIndex.getImmediateChild(INDEX_COL, nextCol);
            }
         }
         
         // Retrieve next index in this class.
         nextIndex = classAst.getImmediateChild(INDEX, nextIndex);
      }
      
      indexMap.put(historical, indexList);
   }
   
   /**
    * Compose and store the set of tables which refer to a database table via
    * a foreign key, given the referent table's P2O class node and its legacy
    * name.  The refering tables are actually stored as a set AST IDs, which
    * is mapped it to the given, legacy table name.
    *
    * @param   historical
    *          Fully qualified, legacy table name.
    * @param   classAst
    *          P2O AST node describing a DMO; the direct parent of all P2O
    *          relation nodes.
    */
   private void initializeRelations(String historical, Aast classAst)
   {
      List<Integer> types = Arrays.asList(
         new Integer[] { new Integer(ONE_TO_MANY), new Integer(ONE_TO_ONE) }
      );
      
      Aast nextRel = null;
      while ((nextRel = classAst.getImmediateChild(types, nextRel)) != null)
      {
         boolean referent = false;
         switch (nextRel.getType())
         {
            case ONE_TO_MANY:
               referent = true;
               break;
            case ONE_TO_ONE:
               referent =
                  !((Boolean) nextRel.getAnnotation(FOREIGN)).booleanValue();
               break;
         }
         
         if (referent)
         {
            // Add AST ID to set of refering tables for this referent table.
            Set<Long> referers = foreignReferenceMap.get(historical);
            if (referers == null)
            {
               referers = new LinkedHashSet<Long>();
               foreignReferenceMap.put(historical, referers);
            }
            referers.add((Long) nextRel.getAnnotation(REFERENCES));
         }
      }
   }
   
   /**
    * Load the persisted, P2O AST associated with the specified database
    * schema.
    *
    * @param   database
    *          Logical name of database schema associated with the P2O
    *          AST to be loaded.
    *
    * @return  P2O AST associated with <code>database</code>.
    *
    * @throws  ConfigurationException
    *          if there is any error loading P2J schema configuration data.
    * @throws  AstException
    *          if there is any error loading the AST from persistence.
    */
   private Aast loadAst(String database)
   throws ConfigurationException,
          AstException
   {
      SchemaConfig config = Configuration.getSchemaConfig();
      File file = config.getSchemaFile(database);
      String path = file.getAbsolutePath();
      path = path.substring(0, path.lastIndexOf(".")) + P2O_POSTFIX;
      path = Configuration.normalizeFilename(path);

      return AstManager.get().loadTree(path);
   }
   
   /**
    * Dump an exhaustive tree of nested table dependencies.  A dependency is
    * a foreign key relation from the dependent table to the "parent" table.
    * The tree is printed to stdout.  Each indent indicates a dependency to
    * the next outer level.
    *
    * @param   historical
    *          Fully qualified, legacy Progress table name.
    * @param   indent
    *          Level of indent to apply when printing this table.
    */
   private void dumpTableDependencies(String historical, int indent)
   {
      StringBuffer buf = new StringBuffer();
      
      StringHelper.repeatChar(' ', indent * 3, buf);
      
      buf.append(historical);
      buf.append(" (");
      buf.append(getJavaName(historical, null));
      buf.append(")");
      
      System.out.println(buf.toString());
      
      Set<Long> referers = foreignReferenceMap.get(historical);
      if (referers != null)
      {
         Iterator<Long> iter = referers.iterator();
         while (iter.hasNext())
         {
            Long id = iter.next();
            Aast classAst = classByIDMap.get(id);
            if (classAst != null)
            {
               StringBuilder hstBuf = new StringBuilder();
               hstBuf.append(classAst.getParent().getText());
               hstBuf.append(".");
               String hist = (String) classAst.getAnnotation(HISTORICAL);
               hstBuf.append(hist);
               dumpTableDependencies(hstBuf.toString(), indent + 1);
            }
         }
      }
   }
   
   /** Container with user-specific data. */
   private static class WorkArea
   {
      /** Map of schema names to P2O lookup instances */
      private final Map<String, P2OLookup> schemaMap = new HashMap<String, P2OLookup>();
      
      /** Set of all known schema names from P2J conversion configuration */
      private Set<String> schemaNames = null;
      
      /** Name of last temp-table P2O AST file to be loaded */
      private String lastTempFilename = null;
   }
   
   /**
    * Test harness for property name lookup feature.  Expects no arguments.
    *
    * @param   args
    *          Not used.
    */
   public static void main(String[] args)
   {
      try
      {
         BufferedReader in = new BufferedReader(
            new InputStreamReader(System.in));
         while (true)
         {
            System.out.print("Enter test data or ':q' to quit:  ");
            String input = in.readLine();
            if (":q".equalsIgnoreCase(input))
            {
               System.out.println("Goodbye.");
               break;
            }
            
            try
            {
//               System.out.println("Qualified java property name:  " +
//                  P2OLookup.javaPropertyName(input, true));
//               System.out.println("Legacy field names:  " +
//                  P2OLookup.legacyFieldNames(input));
//               System.out.println("Is DMO interface ---> " +
//                  P2OLookup.isDMOInterface(input));
//               System.out.println("Is referenced by foreign key ---> " +
//                  P2OLookup.isForeignReference(input));
               System.out.println("Table Dependencies:");
               P2OLookup.dumpTableDependencies(input);
            }
            catch (SchemaException exc)
            {
               System.err.println(exc.getMessage());
            }
         }
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
   }
}
